# -*- coding: utf-8 -*-

import sys
import os.path

import json

from pathlib import Path
import argparse

class PrintConfigAction(argparse.Action):

    @classmethod
    def set_callback_methods(cls, callback_methods):
        cls.callback_methods = callback_methods if callback_methods else []
    
    def __init__(self
                 , option_strings
                 , dest
                 , default=False
                 , required=False
                 , help=None):
        super().__init__(option_strings=option_strings
                         , dest=dest
                         , nargs=0
                         , const=True
                         , default=default
                         , required=required
                         , help=help)                                       

    def __call__(self, parser, namespace, values, option_string=None):
        assert self.callback_methods is not None, 'callback_method cannot be null.'
        if len(self.callback_methods) == 0:
            print('callback_methods must be given to print_conf', file = sys.stderr)
            exit(1)
        for callback_method in self.callback_methods:
            loader = PrintConfig(callback_method.__SECTION__, sys.stdout)
            callback_method(loader)
        exit(0)
            
class PrintConfig:

    def __init__(self, section, out):
        self.section = section
        self.out = out
        print(file = self.out)
        print(file = self.out)
        print('[{}]'.format(section), file = self.out)
        print(file = self.out)

    def add_attr(self, name, default=None, required=False, help=''):
        self.print_attr(name, default, required, help, str)
        
    def add_attr_boolean(self, name, default=False, required=False, help=''):
        self.print_attr(name, default, required, help, bool)

    def add_attr_int(self, name, default=0, required=False, help=''):
        self.print_attr(name, default, required, help, int)

    def add_attr_float(self, name, default=0.0, required=False, help=''):
        self.print_attr(name, default, required, help, float)

    def add_attr_dict(self, name, default={}, required=False, help=''):
        self.print_attr(name, default, required, help, dict, json_data=True)
        
    def add_attr_list(self, name, default=[], required=False, help=''):
        self.print_attr(name, default, required, help, list, json_data=True)
        
    def add_attr_path(self, name, default=None, required=False, help=''):
        self.print_attr(name, default, required, help, Path)

    def print_attr(self, name, default, required, help, typestring, json_data=False):
        if required:
            print('# REQUIRED', file = self.out)
        print('# {} ({})'.format(help, typestring), file = self.out)
        if default is not None:
            print('# {} = {}'.format(name, default if not json_data else json.dumps(default)), file = self.out)
        else:
            print('# {} (no default value)'.format(name), file = self.out)
        if required:
            if default:
                print('{} = {}'.format(name, default), file = self.out)
            else:
                print('{} = '.format(name), file = self.out)
        print('', file = self.out)
        




# -*- coding: utf-8 -*-

import collections
from collections import namedtuple
from recordclass import recordclass

import logging

from ifconf.config import Config

from ifconf.config_print import PrintConfigAction
from ifconf.config_logging import configure_logging

__MAIN_CONFIG__ = None

def config_callback(section = None):
    def _decorator(func):
        if hasattr(func, '__SECTION__'):
            return func
        if func.__module__  == '__main__':
            try:
                func.__MODULE_NAME__ = os.path.splitext(os.path.basename(sys.modules['__main__'].__file__))[0]
            except:
                func.__MODULE_NAME__ = os.path.splitext(os.path.basename(sys.executable))[0]
        else:
            func.__MODULE_NAME__ = func.__module__
        func.__SECTION__ = '_'.join((func.__MODULE_NAME__, str(section) if section else func.__name__))
        return func
    return _decorator

def configure_main(argparser = None
                   , with_default_args = True
                   , config_arg = 'config.ini'
                   , config_path = []
                   , with_config_logging = True
                   , callback_methods = []):
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = Config(argparser)
    callback_methods = callback_methods if hasattr(callback_methods, '__iter__') else [callback_methods]
    callback_methods = [config_callback()(m) for m in callback_methods]
    if with_default_args:
        add_default_argument(__MAIN_CONFIG__.argparser, config_arg)
        PrintConfigAction.set_callback_methods(callback_methods)
    __MAIN_CONFIG__.parse(config_path)
    if with_config_logging:
        configure_logging(__MAIN_CONFIG__)
    for m in callback_methods:
        loader = ConfigLoader.load(m, __MAIN_CONFIG__)
        try:
            loader.configure()
        except Exception as e:
            __MAIN_CONFIG__.err.append('モジュール[{}]の設定取得に失敗しました。エラー：{}'.format(loader.section, e))
    for e in __MAIN_CONFIG__.err:
        __MAIN_CONFIG__.logger.warning(e)
    __MAIN_CONFIG__.logger.info('設定が完了しました。設定ファイル：{}'.format(__MAIN_CONFIG__.config_path))
    return __MAIN_CONFIG__

def configure_main_custom(argparser = None):
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = Config(argparser)
    return __MAIN_CONFIG__

def clear_main_cofig():
    global __MAIN_CONFIG__
    __MAIN_CONFIG__ = None

def configure_module(callback_methods
                     , config = None
                     , immutable = True):
    if config:
        config = config
    elif __MAIN_CONFIG__:
        config = __MAIN_CONFIG__
    else:
        raise RuntimeError('Main configuration must be done before calling configure_module.')
    if hasattr(callback_methods, '__iter__'):
        return [ConfigLoader.load(config_callback()(callback), config).configure(immutable) for callback in callback_methods]
    else:
        return ConfigLoader.load(config_callback()(callback_methods), config).configure(immutable)



def add_default_argument(argparser, config_path = None):
    if config_path is not None:
        argparser.add_argument('-c', '--config'
                               , metavar='PATH'
                               , default=config_path
                               , help='設定ファイルへのパス')
    argparser.add_argument('--verbose', '-v'
                           , action='count'
                           , default=False
                           , help='詳細なデータを出力')
    argparser.add_argument('--current_dir'
                           , default='.'
                           , help='カレントディレクトリを指定して実行')
    argparser.add_argument('--debug'
                           , action="store_true"
                           , default=False
                           , help='デバグモードで実行')
    argparser.add_argument('--debug_file'
                           , metavar='FILE'
                           , help='デバグモードで実行しファイル出力')
    argparser.add_argument('--print_conf'
                           , action=PrintConfigAction
                           , default=False
                           , help='設定ファイルの内容を出力')
    return argparser

    

class ConfigLoader:
    
    @classmethod
    def load(cls, callback_method, config):
        assert callback_method is not None, 'callback_method cannot be null.'
        assert hasattr(callback_method, '__MODULE_NAME__'), 'MODULE_NAME must be set for callback_method.'
        assert hasattr(callback_method, '__SECTION__'), 'SECTION must be set for callback_method.'
        loader = ConfigLoader(callback_method.__SECTION__, callback_method.__MODULE_NAME__, config)
        callback_method(loader)
        return loader
        
    def __init__(self, section, module_name, config):
        assert section is not None, 'section cannot be null.'
        assert module_name is not None, 'module_name cannot be null.'
        assert config is not None, 'config cannot be null.'
        assert type(config) is Config, 'invalid config type:[{}]'.format(type(config))
        self.section = section
        self.main_config = config
        self.names = ['main_config', 'logger']
        self.values = [lambda config: config, lambda config: logging.getLogger(module_name)]

    def configure(self, immutable):
        assert self.main_config is not None, 'main_config cannot be null.'
        ntp = namedtuple(self.section.replace('.','_'), self.names) if immutable else recordclass(self.section.replace('.','_'), self.names) 
        args = [f(self.main_config) for f in self.values]
        conf = ntp(*args)
        conf.logger.debug(conf)
        return conf

    def add_attr(self, name, default=None, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr(self.section, name, default, required))

    def add_attr_boolean(self, name, default=False, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_boolean(self.section, name, default, required))

    def add_attr_int(self, name, default=0, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_int(self.section, name, default, required))

    def add_attr_float(self, name, default=0.0, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_float(self.section, name, default, required))

    def add_attr_dict(self, name, default={}, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_dict(self.section, name, default, required))
        
    def add_attr_list(self, name, default=[], required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_list(self.section, name, default, required))
        
    def add_attr_path(self, name, default=None, required=False, help=''):
        self.names.append(name)
        self.values.append(lambda conf: conf.get_attr_path(self.section, name, default, required))





# -*- coding: utf-8 -*-

import logging
import logging.config
from logging.handlers import RotatingFileHandler

logging.NOTICE = 25
logging.NOTICE_LEVEL_NAME = "NOTICE"

logging.addLevelName(logging.NOTICE, logging.NOTICE_LEVEL_NAME)
def notice(self, message, *args, **kws):
    self._log(logging.NOTICE, message, args, **kws) 
logging.Logger.notice = notice

def configure_logging(conf
                      , log_level=logging.INFO
                      , log_file=""
                      , log_max_bytes = 1024*1024*16
                      , log_backup_count = 99
                      , log_encoding = 'UTF-8'
                      , fmt = '%(asctime)s [%(name)s] [%(levelname)s] %(message)s'):
    conf.verbose = conf.args.verbose if hasattr(conf.args, 'verbose') else False
    conf.log_level = log_level
    conf.log_level_name = logging.getLevelName(log_level)
    conf.log_file = log_file
    conf.log_max_bytes = log_max_bytes
    conf.log_backup_count = log_backup_count
    conf.log_encoding = log_encoding
    conf.formatter = logging.Formatter(fmt)

    if [s for s in ('loggers', 'handlers', 'formatters') if conf.parser.has_section(s)]:
        logging.config.fileConfig(conf.parser)
        conf.logger = logging.getLogger()
    elif conf.parser.has_section('logging'):
        section = conf.parser['logging']
        conf.verbose = section.get('verbose', conf.verbose)
        conf.log_level_name = section.get('level', 'INFO')
        conf.log_level = getattr(logging, conf.log_level_name.upper(), conf.log_level)
        conf.log_file = section.get('log_file', conf.log_file)
        conf.err_file = section.get('err_file', '')
        conf.log_encoding = section.get('encoding', conf.log_encoding)
        bytes_str = section.get("max_bytes", '').replace('(','').replace(')','').replace('=','')
        conf.log_max_bytes = eval(bytes_str) if bytes_str else conf.log_max_bytes
        conf.log_backup_count = int(section.get('backup_count', conf.log_backup_count))
        conf.logger = load_logging(conf)
    else:
        conf.logger = load_logging(conf)

def load_logging(conf, name=None):
    if name:
        logger = logging.getLogger(name)
    else:
        logger = logging.getLogger()
    logger.setLevel(conf.log_level)

    if hasattr(conf.args, 'debug') and conf.args.debug:
        handler = logging.StreamHandler()
        handler.setFormatter(conf.formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
            
    if hasattr(conf.args, 'debug_file') and conf.args.debug_file:
        try:
            handler = FileHandler(conf.args.debug_file, encoding = conf.log_encoding)
            handler.setFormatter(conf.formatter)
            handler.setLevel(logging.DEBUG)
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)
        except Exception as e:
            conf.err.append('DEBUG FILEログ設定に失敗しました。ファイル：[{}] エラー：[{}]'.format(conf.args.debug_file, e))
    if hasattr(conf.args, 'log_file') and conf.log_file:
        try:
            handler = RotatingFileHandler(conf.log_file, maxBytes = conf.log_max_bytes, backupCount = conf.log_backup_count, encoding = conf.log_encoding)
            handler.setFormatter(conf.formatter)
            logger.addHandler(handler)
        except Exception as e:
            conf.err.append('ログ設定に失敗しました。ファイル：[{}] エラー：[{}]'.format(conf.log_file, e))
    if hasattr(conf.args, 'err_file') and conf.err_file != conf.log_file:
        try:
            handler = RotatingFileHandler(conf.err_file, maxBytes = conf.log_max_bytes, backupCount = conf.log_backup_count, encoding = conf.log_encoding)
            handler.setFormatter(conf.formatter)
            handler.setLevel(logging.ERROR)
            logger.addHandler(handler)
        except Exception as e:
            conf.err.append('エラーログ設定に失敗しました。ファイル：[{}] エラー：[{}]'.format(conf.err_file, e))
    if len(logger.handlers) == 0:
        handler = logging.StreamHandler()
        handler.setFormatter(conf.formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        logger.warning("ログの出力先が標準エラー出力に設定されました。")
        
    if conf.verbose:
        logger.setLevel(logging.DEBUG)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUGモードが有効になりました。")
    elif logger.isEnabledFor(logging.INFO):
        logger.info("ログ設定が有効になりました。INFOログが出力されます。")
    elif logger.isEnabledFor(logging.NOTICE):
        logger.notice("ログ設定が有効になりました。NOTICEログが出力されます。")
    elif logger.isEnabledFor(logging.WARN):
        logger.warning("ログ設定が有効になりました。WARNINGログが出力されます。")
        
    return logger

